/*
 * Enhydra Java Application Server Project The contents of this file are subject to the Enhydra Public License Version 1.1 (the "License"); you may not use this
 * file except in compliance with the License. You may obtain a copy of the License on the Enhydra web site ( http://www.enhydra.org/ ). Software distributed
 * under the License is distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the specific terms
 * governing rights and limitations under the License. The Initial Developer of the Enhydra Application Server is Lutris Technologies, Inc. The Enhydra
 * Application Server and portions created by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc. All Rights Reserved. Contributor(s): $Id:
 * BMByteSearch.java,v 1.1 2005/07/01 15:11:34 pyropunk Exp $
 */
package com.lutris.util;

/**
 * Implements the Boyer-Moore pattern matching algorithm for a given byte pattern. This object implements searches on signed byte arrays. The algorithm was
 * obtained from "Computer Algorithms - Introduction to Design and Analysis, Second Edition" by Sara Baase.
 * 
 * @author <a href="http://www.enhydra.org/">Enhydra</a>
 * @version 1
 */
public class BMByteSearch {
  /**
   * Byte array, beginning at index 1 (for algorithmic convenience), that contains the intended search pattern data.
   */
  private byte[] P;
  /**
   * The length of the search pattern.
   */
  private int m;
  /**
   * Table of jump distances for each mismatched character in the alphabet for a given search pattern. Must be recomputed for each new pattern.
   */
  private int[] charJump;
  /**
   * Table of partial suffix match jump distances for a given pattern. Must be recomputed for each new pattern.
   */
  private int[] matchJump;

  /**
   * Creates a precomputed Boyer-Moore byte string search object from the given pattern. The unicode characters in <code>pattern</code> are truncated if
   * greater than 255, and converted in twos-complement fashion, to appropriate negative byte values, if necessary. This method is provided as a convenience for
   * searching for patterns within 8 bit byte strings composed of character data.
   * 
   * @param pattern The pattern create this object for.
   */
  public BMByteSearch(final String pattern) {
    this(genPatternFromCharArray(pattern.toCharArray()));
  }

  /**
   * Creates a precomputed Boyer-Moore byte string search object from the given pattern.
   * 
   * @param pattern Binary pattern to search for.
   */
  public BMByteSearch(final byte[] pattern) {
    this(pattern, 0, pattern.length);
  }

  /**
   * Creates a precomputed Boyer-Moore byte string search object from a portion of the given pattern array.
   * 
   * @param pattern Byte array containing a pattern to search for.
   * @param offset Offset to beginning of search pattern.
   * @param length Length of the search pattern.
   */
  public BMByteSearch(final byte[] pattern, final int offset, final int length) {
    genPatternFromByteArray(pattern, offset, length);
    computeJumps();
    computeMatchJumps();
  }

  /**
   * Compares two integers and returns the lesser value.
   * 
   * @param i1 First integer to compare.
   * @param i2 Second integer to compare.
   * @return The lesser of <code>i1</code> or <code>i2</code>.
   */
  private static final int min(final int i1, final int i2) {
    return i1 < i2 ? i1 : i2;
  }

  /**
   * Compares two integers and returns the greater value.
   * 
   * @param i1 First integer to compare.
   * @param i2 Second integer to compare.
   * @return The greater of <code>i1</code> or <code>i2</code>.
   */
  private static final int max(final int i1, final int i2) {
    return i1 > i2 ? i1 : i2;
  }

  /**
   * Generates the pattern byte string <code>P</code> from a portion of another byte string.
   * 
   * @param bytes The byte string from which to extract the pattern.
   * @param off The array index within <code>bytes</code> from which to extract the pattern.
   * @param length The number of characters to extract from <code>bytes</code> into the pattern.
   */
  private final void genPatternFromByteArray(final byte[] bytes, final int off, final int length) {
    m = length;
    // 31.03.2003. patch // P = new byte[length];
    P = new byte[length + 1];
    System.arraycopy(bytes, off, P, 1, length);
    /**
     * 2005.09.21 fix by pyropunk
     * 
     * <pre>
     * int i;
     * int j;
     * for (i = 1, j = off; (i &lt;= length); i++, j++) {
     *   P[i] = bytes[j];
     * }
     * </pre>
     */
  }

  /**
   * Generates the pattern byte string <code>P</code> from a character array. The signed unicode characters are truncated to 8 bits, and converted into signed
   * byte values. Characters between 128 and 255 are converted to their signed negative counterpart in twos-complement fashion by subtracting 256.
   * 
   * @param chars Unsigned unicode character array to turn into a signed byte array.
   * @return a <code>byte[]</code> translation of the original <code>char[]</code>
   */
  private static final byte[] genPatternFromCharArray(final char[] chars) {
    final int m_loc = chars.length;
    final byte[] P_loc = new byte[m_loc + 1];
    for (int i = 1; i <= m_loc; i++) {
      if (chars[i - 1] > 127) {
        P_loc[i] = (byte)(chars[i - 1] - 256 & 0xff);
      } else {
        P_loc[i] = (byte)(chars[i - 1] & 0xff);
      }
    }
    return P_loc;
  }

  /**
   * Initializes the per-character jump table <code>charJump</code> as specified by the Boyer-Moore algorithm.
   */
  private final void computeJumps() {
    charJump = new int[256];
    for (int i = 0; i < 255; i++) {
      charJump[i] = m;
    }
    for (int k = 1; k <= m; k++) {
      charJump[P[k] + 128] = m - k;
    }
  }

  /**
   * Computes a partial-match jump table that skips over partially matching suffixes.
   */
  private void computeMatchJumps() {
    int k = 0;
    int q = 0;
    int qq = 0;
    final int mm;
    final int[] back = new int[(m + 2)];
    final int mq;
    matchJump = new int[m + 2];
    mm = m << 1;
    for (k = 1; k <= m; k++) {
      matchJump[k] = mm - k;
    }
    k = m;
    q = m + 1;
    while (k > 0) {
      final int mk = m - k;
      back[k] = q;
      while (q <= m && P[k] != P[q]) {
        matchJump[q] = min(matchJump[q], mk);
        q = back[q];
      }
      k--;
      q--;
    }
    mq = m + q;
    for (k = 1; k <= q; k++) {
      matchJump[k] = min(matchJump[k], mq - k);
    }
    qq = back[q];
    while (q <= m) {
      while (q <= qq) {
        matchJump[q] = min(matchJump[q], qq - q + m);
        q++;
      }
      qq = back[qq];
    }
  }

  /**
   * Returns the length of the pattern for this searcher.
   * 
   * @return The search pattern length.
   */
  public final int getPatternLength() {
    return m;
  }

  /**
   * Search for the previously pre-compiled pattern string in an array of bytes. This method uses the Boyer-Moore pattern search algorithm.
   * 
   * @param byteString Array of bytes in which to search for the pattern.
   * @return The array index where the pattern begins in the string, or <code>-1</code> if the pattern was not found.
   */
  public int search(final byte[] byteString) {
    return search(byteString, 0, byteString.length);
  }

  /**
   * Search for the previously pre-compiled pattern string in an array of bytes. This method uses the Boyer-Moore pattern search algorithm.
   * 
   * @param byteString Array of bytes in which to search for the pattern.
   * @param offset The the index in <code>byteString</code> where the search is to begin.
   * @param length The number of bytes to search in <code>byteString</code>.
   * @return The array index where the pattern begins in the string, or <code>-1</code> if the pattern was not found.
   */
  public int search(final byte[] byteString, final int offset, final int length) {
    int j = 0;
    int k = 0;
    int len = 0;
    j = m + offset;
    k = m;
    len = min(byteString.length, offset + length);
    while (j <= len && k > 0) {
      if (byteString[j - 1] == P[k]) {
        j--;
        k--;
      } else {
        j += max(charJump[byteString[j - 1] + 128], matchJump[k]);
        k = m;
      }
    }
    if (k != 0) {
      // No match.
      j = -1;
    }
    return j;
  }

  /**
   * @see java.lang.Object#toString()
   */
  @Override
  public String toString() {
    return this.getClass().getName();
  }
}
